/*======================================================================================

	Program:		Range Imaging Toolkit (RITK)
	Module:			ritkCuda3DArrayImportImageContainer.h
	Version:		$Revision: 2774 $
	Date:			$Date: 2012-03-13 10:52:35 +0100 (Di, 13 Mrz 2012) $

	Copyright (c) 2010-2011
	Pattern Recognition Lab, 
    University Erlangen-Nuremberg, Germany
    All rights reserved.

	This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
	See RITKCopyright.txt for more information.

	http://www5.informatik.uni-erlangen.de/research/software/ritk

========================================================================================*/

#ifndef CUDA3DARRAYIMPORTIMAGECONTAINER_H__
#define CUDA3DARRAYIMPORTIMAGECONTAINER_H__

#include "ritkCudaImportImageContainer.h"

namespace ritk
{

/**	@ingroup	Cuda
 *	@class		Cuda3DArrayImportImageContainer
 *	@author		Jakob Wasza
 *	@brief		Memory container for 3D CUDA arrays
 *
 *	@details
 *	This class provides methods for CUDA 3D array management
 */
template<typename TElementIdentifier, typename TElement>
class Cuda3DArrayImportImageContainer : public CudaImportImageContainer<TElementIdentifier,TElement,3>
{
public:
	/**	@name	Std typedefs */
	//@{
	typedef Cuda3DArrayImportImageContainer								Self;
	typedef CudaImportImageContainer<TElementIdentifier,TElement,3>		Superclass;
	typedef itk::SmartPointer<Self>										Pointer;
	typedef itk::SmartPointer<const Self>								ConstPointer;
	//@}

	/** @name	Generic import image container types */
	//@{
	typedef typename Superclass::GenericImportImageContainerType		GenericImportImageContainerType;
	typedef typename Superclass::ImportImageContainerType				ImportImageContainerType;
	//@}

	/**	@name	Template parameters */
	//@{
	typedef TElementIdentifier  ElementIdentifier;
	typedef TElement            Element;
	//@}

	//@}

	/// Object creation
	itkNewMacro(Self);

	/// RTTI
	itkTypeMacro(Cuda3DArrayImportImageContainer, CudaImportImageContainer);

	/**	@name	Host/device synchronization.*/
	//@{
	virtual void SynchronizeDevice();
	virtual void SynchronizeHost();
	//@}

	/**	@name	Deep copy */
	//@{

	/**	@brief	Deep copy the buffer
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	Copy data from an itk::ImportImageContainer, however, this class' implementation only addresses copying device memory,
	 *	host memory copy will be performed by the baseclass that eventually also allocates memory (host AND device). \n
	 *	This special CUDA container will copy the CUDA memory if a CUDAImportImageContainer can be derived from the parameter.
	 *	This class' implementation will automatically determine and execute the proper copy scheme (i.e. Array -> Regular, ...).
	 *	Note that copying device memory will be performed if and only if the device modified time is equal or more up-to-date 
	 *	than the host memory. 
	 */
	virtual void DeepCopy(ImportImageContainerType *Container);


	/**	@brief	Deep copy from a container
	 *	@param	Container	The container providing the source
	 *
	 *	@details
	 *	This method will simply remove the const qualifier (const_cast) of the input argument and pass it to the non-const DeepCopy(ImportImageContainerType *) method.
	 *	Note that a non-const input is necessary due to the synchronization mechanism. However, it is granted, that the memory (host or device) that is up-to-data
	 *	will NOT be modified!
	 */
	virtual void DeepCopy(const ImportImageContainerType *Container);

	//@}

	/// Copy to a mapped resource
	virtual void CopyToMappedResource(void *d_ptr, size_t size) const;


	/**	@name	Pointer to the wrapped CUDA memory array. Non-const version will modify the device MTime! */
	//@{
	cudaArray *GetCudaMemoryArray();
	const cudaArray *GetCudaMemoryArray() const;
	//@}

	/**	@name Get the channel descriptor of the wrapped array */
	//@{
	cudaChannelFormatDesc &GetChannelDesc();
	const cudaChannelFormatDesc &GetChannelDesc() const;
	//@}


protected:
	/// Constructor
	Cuda3DArrayImportImageContainer();
	/// Destructor
	virtual ~Cuda3DArrayImportImageContainer();

	/// Print
	virtual void PrintSelf(std::ostream& os, itk::Indent indent) const;

	/// Allocate the buffer. Is delegated to the superclass.
	virtual Element* AllocateElements(ElementIdentifier size) const;

	/// Free the buffer. Is delegated to the superclass
	virtual void DeallocateManagedMemory();

	/**	@name	Methods that have to be executed in the global CUDA context */
	//@{
	CudaConstMethodMacro(Cuda3DArrayImportImageContainer, AllocateCudaMemory);
	CudaMethodMacro(Cuda3DArrayImportImageContainer, DeallocateCudaMemory);
	CudaMethodMacro(Cuda3DArrayImportImageContainer, CopyDeviceToHost);
	CudaMethodMacro(Cuda3DArrayImportImageContainer, CopyHostToDevice);
	Superclass *m_TmpCudaContainer;
	CudaMethodMacro(Cuda3DArrayImportImageContainer, DeepCopyCudaMemory);
	//@}

protected:
	/// List holding the memory. This is due to the (de)allocate scheme in itk. Always delete the front and push new memory
	mutable std::list<std::pair<cudaArray*,cudaChannelFormatDesc> > m_CudaArray;

private:
	/// purposely not implemented
	Cuda3DArrayImportImageContainer(const Self&); 
	/// purposely not implemented
	void operator=(const Self&);
};

/** @name Template instantiations */
//@{
typedef Cuda3DArrayImportImageContainer<unsigned long,float> Cuda3DArrayImportImageContainerF;
//@}

} // ritk


#endif // CUDA3DARRAYIMPORTIMAGECONTAINER_H__